import inspect
import os
import psycopg2
import pandas as pd
from psycopg2 import sql
from typing import Any, Optional
import dotenv


class PostgresConnection:
    def __init__(
        self,
        host: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        database: Optional[str] = None,
    ):
        if os.path.isfile('.env'):
            dotenv.load_dotenv()

        self.host = os.environ['postgres_host'] if host is None else host
        self.username = os.environ['postgres_username'] if username is None else username
        self.password = os.environ['postgres_password'] if password is None else password
        self.database = os.environ['postgres_database'] if database is None else database

        if self.host is None or self.database is None or self.username is None or self.password is None:
            raise RuntimeError('No env. variable set or .env file given while one of the parameters is missing.')

        self.connection: Optional[Any] = None

    def __enter__(self):
        calling_function = inspect.stack()[2][3]

        self.connection = psycopg2.connect(
            host=self.host,
            dbname=self.database,
            user=self.username,
            password=self.password,
            application_name=f'{calling_function}()',
        )

        self.connection.autocommit = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()

    def _check_connection_exists(self):
        if self.connection is None or self.connection.closed:
            raise RuntimeError('Connection not made yet.')

    def _execute_or_fetch_query_on_db(self, query: sql.Composable, fetch: bool = False) -> Optional[list]:
        self._check_connection_exists()
        cursor = self.connection.cursor()  # type:ignore

        # print(f"Executing query: ({query.as_string(connection)})")
        cursor.execute(query)
        # print(f"Succesfully executed query: ({cursor.query.decode()})")

        if fetch:
            records = cursor.fetchall()
            # print('Fetched records')
        else:
            records = None

        cursor.close()

        return records

    def fetch_with_query_on_db(self, query: sql.Composable) -> list:
        return self._execute_or_fetch_query_on_db(query, fetch=True)  # type:ignore

    def execute_query_on_db(self, query: sql.Composable) -> list:
        return self._execute_or_fetch_query_on_db(query, fetch=False)  # type:ignore

    def check_if_table_exists(self, schema_name: str, table_name: str) -> bool:
        exists_query = sql.SQL(
            """select exists (
                        select from information_schema.tables
                        where  table_schema = {schema_name}
                        and    table_name   = {table_name}
            )"""
        ).format(
            schema_name=sql.Literal(schema_name),
            table_name=sql.Literal(table_name),
        )
        exists = self.fetch_with_query_on_db(exists_query)[0][0]
        return exists

    def create_table_if_not_exists(self, schema_name: str, table_name: str, table_columns: sql.SQL, connection: Optional[Any] = None):
        table_exists_query = sql.SQL('select exists( select * FROM pg_catalog.pg_tables WHERE tablename = {table_name} and schemaname = {schema_name})').format(
            table_name=sql.Literal(table_name), schema_name=sql.Literal(schema_name)
        )
        table_exists = self.fetch_with_query_on_db(table_exists_query)[0][0]

        if not table_exists:
            create_table_query = sql.SQL('create table if not exists {schema_name}.{table_name} ({table_columns})').format(
                schema_name=sql.Identifier(schema_name),
                table_name=sql.Identifier(table_name),
                table_columns=table_columns,
            )
            self.execute_query_on_db(create_table_query)

    def insert_df(self, df: pd.DataFrame, schema: str, table: str) -> None:
        self._check_connection_exists()
        cursor = self.connection.cursor()  # type:ignore

        n_columns = df.shape[1]
        insert_string_part = ','.join(['%s'] * n_columns)
        insert_query = f'INSERT INTO {schema}.{table} VALUES ({insert_string_part}) ON CONFLICT do nothing'

        cursor.executemany(insert_query, df.values.tolist())
        self.connection.commit()  # type:ignore

        print('Data inserted successfully.')



import inspect
import os
import psycopg2
import pandas as pd
from psycopg2 import sql
from typing import Any, Optional
import datetime as dt
import dotenv

from .postgres_connection import PostgresConnection  # noqa F402


def get_connection(
    host: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    database: Optional[str] = None,
):
    if os.path.isfile('.env'):
        dotenv.load_dotenv()

    host = os.environ['postgres_host'] if host is None else host
    username = os.environ['postgres_username'] if username is None else username
    password = os.environ['postgres_password'] if password is None else password
    database = os.environ['postgres_database'] if database is None else database

    if host is None or database is None or username is None or password is None:
        raise RuntimeError('No env. variable set or .env file given while one of the parameters is missing.')

    calling_function = inspect.stack()[2][3]

    connection = psycopg2.connect(
        host=host,
        dbname=database,
        user=username,
        password=password,
        application_name=f'{calling_function}()',
    )

    connection.autocommit = True
    return connection


def _execute_or_fetch_query_on_db(query: sql.Composable, fetch: bool = False, connection: Optional[Any] = None) -> Optional[list]:
    # TODO maakt nu iedere keer een nieuwe connectie niet heel nice
    if connection is None:
        close_connection = True
        connection = get_connection()
    else:
        close_connection = False

    connection.autocommit = True
    # print("Created connection")
    cursor = connection.cursor()

    # print(f"Executing query: ({query.as_string(connection)})")
    cursor.execute(query)
    # print(f"Succesfully executed query: ({cursor.query.decode()})")

    if fetch:
        records = cursor.fetchall()
        # print('Fetched records')
    else:
        records = None

    cursor.close()
    if close_connection:
        connection.close()

        if connection.closed:
            # print("Closed connection")
            pass
        else:
            print('Connection not closed yet...')
    else:
        pass
        # print("Warning, not closing connection!")

    return records


def fetch_with_query_on_db(query: sql.Composable, connection: Optional[Any] = None) -> list:
    return _execute_or_fetch_query_on_db(query, fetch=True, connection=connection)  # type:ignore


def execute_query_on_db(query: sql.Composable, connection: Optional[Any] = None) -> list:
    return _execute_or_fetch_query_on_db(query, fetch=False, connection=connection)  # type:ignore


def check_if_table_exists(schema_name: str, table_name: str, connection: Optional[Any] = None) -> bool:
    exists_query = sql.SQL(
        """select exists (
                    select from information_schema.tables
                    where  table_schema = {schema_name}
                    and    table_name   = {table_name}
                )"""
    ).format(
        schema_name=sql.Literal(schema_name),
        table_name=sql.Literal(table_name),
    )
    exists = fetch_with_query_on_db(exists_query, connection)[0][0]
    return exists


def create_table_if_not_exists(schema_name: str, table_name: str, table_columns: sql.SQL, connection: Optional[Any] = None):
    table_exists_query = sql.SQL('select exists( select * FROM pg_catalog.pg_tables WHERE tablename = {table_name} and schemaname = {schema_name})').format(
        table_name=sql.Literal(table_name), schema_name=sql.Literal(schema_name)
    )
    table_exists = fetch_with_query_on_db(table_exists_query)[0][0]

    if not table_exists:
        create_table_query = sql.SQL('create table if not exists {schema_name}.{table_name} ({table_columns})').format(
            schema_name=sql.Identifier(schema_name),
            table_name=sql.Identifier(table_name),
            table_columns=table_columns,
        )
        execute_query_on_db(create_table_query, connection)


def insert_df(df: pd.DataFrame, schema: str, table: str) -> None:
    n_columns = df.shape[1]

    insert_string_part = ','.join(['%s'] * n_columns)

    insert_query = f'INSERT INTO {schema}.{table} VALUES ({insert_string_part}) ON CONFLICT do nothing'

    conn = get_connection()
    cursor = conn.cursor()
    cursor.executemany(insert_query, df.values.tolist())
    conn.commit()
    conn.close()
    print('Data inserted successfully.')


def get_most_recent_time_value(time_column: str, schema: str, table: str) -> dt.datetime:
    """
    Returns the most recent timestamp from the time column from the table, of the table does not exist,
    it returns 1st January 2023
    """
    database_newest_date = dt.datetime(0, 0, 0, tzinfo=dt.timezone.utc)
    if check_if_table_exists(schema, table):
        newest_entry_date_query = sql.SQL('select {time_column_name} from {schema}.{table} order by {time_column_name} desc limit 1').format(
            schema=sql.Identifier(schema),
            table=sql.Identifier(table),
            time_column_name=sql.Identifier(time_column),
        )
        result = fetch_with_query_on_db(newest_entry_date_query)
        if len(result) > 0:
            database_newest_date = result[0][0]

    return database_newest_date



import uuid
import pandas as pd
import pytest
from psycopg2 import sql

from src.postgreasy import PostgresConnection


def test_connect():
    with PostgresConnection() as conn:
        x = conn.fetch_with_query_on_db('select 1')
        print(x)
        assert x == [(1,)]


def test_disconnect():
    with PostgresConnection() as conn:
        x = conn.fetch_with_query_on_db('select 1')
        assert x == [(1,)]

    with pytest.raises(RuntimeError):
        conn.fetch_with_query_on_db('select 1')


def test_insert():
    df = pd.DataFrame({'x': [1, 2, 3], 'y': [5, 28, 8]})
    test_table = f'test_insert_{str(uuid.uuid4())[:4]}'
    with PostgresConnection(database='ovviadatatransaction_test') as conn:
        conn.create_table_if_not_exists('public', test_table, sql.SQL('x int, y int'))
        conn.insert_df(df, 'public', test_table)
        records = conn.fetch_with_query_on_db(sql.SQL('select * from {test_table}').format(test_table=sql.Identifier(test_table)))
        conn.execute_query_on_db(sql.SQL('drop table {test_table}').format(test_table=sql.Identifier(test_table)))
        print(records)
        assert records == [(1, 5), (2, 28), (3, 8)]
